from LMS.Common.LMS_Block import LMS_Block
from LMS.Stream.Reader import Reader


class SYL3:
    """A class that represents a SYL3 block in a MSBP file.

    https://github.com/kinnay/Nintendo-File-Formats/wiki/MSBP-File-Format#syl3-block"""

    def __init__(self):
        self.block: LMS_Block = LMS_Block()
        self.styles: list[dict] = []

    def read(self, reader: Reader) -> None:
        """Reads the SYL3 block from a stream.

        :param `reader`: A Reader object."""
        self.block.read_header(reader)

        style_count = reader.read_uint32()
        # Read the styles
        for _ in range(style_count):
            style = {}
            style["region_width"] = reader.read_uint32()
            style["line_num"] = reader.read_uint32()
            style["font_index"] = reader.read_uint32()
            style["base_color_index"] = reader.read_uint32()

            self.styles.append(style)

        self.block.seek_to_end(reader)



from LMS.Stream.Reader import Reader
from LMS.Stream.Writer import Writer


def read_System_Ruby(reader: Reader):
    data = {}
    data["rt"] = reader.read_len_prefixed_utf16_string()

    return data


def write_System_Ruby(writer: Writer, data: dict):
    writer.write_len_prefixed_utf16_string(data["rt"])

    return


def read_System_Font(reader: Reader):
    data = {}
    data["face"] = reader.read_len_prefixed_utf16_string()

    return data


def write_System_Font(writer: Writer, data: dict):
    writer.write_len_prefixed_utf16_string(data["face"])

    return


def read_System_Size(reader: Reader):
    data = {}
    data["percent"] = reader.read_uint16()

    return data


def write_System_Size(writer: Writer, data: dict):
    writer.write_uint16(int(data["percent"]))

    return


def read_System_Color(reader: Reader):
    data = {}
    data["r"] = reader.read_uint8()

    data["g"] = reader.read_uint8()

    data["b"] = reader.read_uint8()

    data["a"] = reader.read_uint8()

    return data


def write_System_Color(writer: Writer, data: dict):
    writer.write_uint8(int(data["r"]))

    writer.write_uint8(int(data["g"]))

    writer.write_uint8(int(data["b"]))

    writer.write_uint8(int(data["a"]))

    return


def read_System_PageBreak(reader: Reader):
    data = {}
    return data


def write_System_PageBreak(writer: Writer, data: dict):
    return

base_preset = { "data": {
	0: {
	'name': 'System', 
	'tags': [
		{'name': 'Ruby', 'read_function' : read_System_Ruby, 'write_function': write_System_Ruby},
		{'name': 'Font', 'read_function' : read_System_Font, 'write_function': write_System_Font},
		{'name': 'Size', 'read_function' : read_System_Size, 'write_function': write_System_Size},
		{'name': 'Color', 'read_function' : read_System_Color, 'write_function': write_System_Color},
		{'name': 'PageBreak', 'read_function' : read_System_PageBreak, 'write_function': write_System_PageBreak},
	]},
}}


from LMS.Stream.Reader import Reader
from LMS.Stream.Writer import Writer

class LMS_Block:
    """A class that represents a block in a LMS file."""
    def __init__(self):
        self.magic: str = None 
        self.size: int = None 
        self.data_start: int = None 
 
    def read_header(self, reader: Reader) -> None:
        """Reads the block header from a stream
        
        `reader`: A Reader object."""
        self.magic = reader.read_string_len(4) 
        self.size = reader.read_uint32()
        reader.skip(8)
        self.data_start = reader.tell()
    
    def get_item_offsets(self, reader: Reader, item_count: int) -> list[int]:
        """Returns the absolute offsets to items in a block.
        
        :param `reader`: A Reader object.
        :param `item_count`: Amount of items."""
        return [reader.read_uint32() + self.data_start for _ in range(item_count)]
    
    def seek_to_end(self, stream: Reader | Writer) -> None:
        """Seeks to the end of a block passed the AB padding
        
        `stream`: A Writer or Reader object."""
        stream.seek(self.data_start)
        end = self.size + 16 - self.size % 16
        stream.seek(end, 1)

    def write_header(self, writer: Writer) -> None:
        """Writes the block to a stream.
        
        `writer`: A Writer object."""
        writer.write_string(self.magic)
        writer.write_uint32(0)
        writer.write_bytes(b"\x00" * 8)
        self.data_start = writer.tell()

    def write_ab_padding(self, writer: Writer) -> None:
        """Writes the ab padding after a block.
        
        `writer`: A Writer object."""
        
        remainder = 16 - self.size % 16
        if remainder == 16:
            return 0
        
        writer.write_bytes(b"\xAB" * remainder)
        return remainder

    def write_size(self, writer: Writer):
        """Writes the size of the block.
        
        `writer`: A Writer object."""
        writer.seek(self.data_start - 12)
        writer.write_uint32(self.size)

    def write_end_data(self, writer: Writer):
        """Writes the ab padding, size, and seeks to the end of a block.
        
        `writer`: A Writer object."""
        offset = writer.tell()
        remainder = self.write_ab_padding(writer)
    
        self.write_size(writer)
        writer.seek(offset)
        
        if remainder == 0:
            return
        
        writer.write_bytes(b"\xAB" * remainder)
        self.seek_to_end(writer)
   


from LMS.Common.LMS_Block import LMS_Block
from LMS.Stream.Reader import Reader


class TGG2:
    """A class that represents a TGG2 block in a MSBT file.

    https://github.com/kinnay/Nintendo-File-Formats/wiki/MSBP-File-Format#tgg2-block"""

    def __init__(self):
        self.block: LMS_Block = LMS_Block()
        self.groups: list[str] = []

    def read(self, reader: Reader) -> None:
        """Reads the TGG2 block from a stream.

        :param `reader`: A Reader object."""
        self.block.read_header(reader)

        group_count = reader.read_uint16()
        reader.skip(2)
        # Read the tag groups
        for offset in self.block.get_item_offsets(reader, group_count):
            group = {}
            reader.seek(offset)

            group["tag_count"] = reader.read_uint16()
            group["tag_indexes"] = [
                reader.read_uint16() for _ in range(group["tag_count"])
            ]
            group["name"] = reader.read_string_nt()
            self.groups.append(group)

        self.block.seek_to_end(reader)



from LMS.Common.LMS_Block import LMS_Block
from LMS.Stream.Reader import Reader
from LMS.Stream.Writer import Writer
from LMS.Project.MSBP import MSBP
from LMS.Common.LMS_Enum import LMS_Types

from LMS.Common.Base_Preset import base_preset

import re


class TXT2:
    """A class that represents a TXT2 block in a MSBT file.

    https://github.com/kinnay/Nintendo-File-Formats/wiki/MSBT-File-Format#txt2-block"""

    def __init__(self):
        self.block: LMS_Block = LMS_Block()
        self.messages: list[str] = []
        self.preset: list[dict] = base_preset
        self.preset_set = False

    def tag_encoded(self, tag: str) -> bool:
        """Returns if a tag is encoded or not.

        :param `tag`: The tag to check."""
        return "." in tag[: tag.find(":")]

    def get_decoded_tag_information(self, tag: str) -> dict:
        """Returns a dictionary of all the information on a tag if its decoded.

        :param `tag`: The tag to get
        the infomation for."""
        result = {}
        if not self.tag_encoded(tag):
            group_name = tag[1 : tag.find(":")]
            tag_name = tag[tag.find(":") + 1 : tag.rfind(":")]

            result["group_name"] = group_name
            result["tag_name"] = tag_name
            dictionary_parameters = dict(
                re.findall(r'(\w+)\s*=\s*["\']([^"\']+)["\']', tag)
            )

            result["parameters"] = dictionary_parameters
            return result

    def get_tag_info(self, tag: str) -> dict:
        """Returns a dictionary of all the information on a tag assuming its decoded.
        
        :param `tag`: The tag to get the information for."""
        result = {}

        if self.tag_encoded(tag):
            return

        group_name = tag[1 : tag.find(":")]
        tag_name = tag[tag.find(":") + 1 : tag.rfind(":")]

        result["group_name"] = group_name
        result["tag_name"] = tag_name
        dictionary_parameters = dict(
            re.findall(r'(\w+)\s*=\s*["\']([^"\']+)["\']', tag)
        )

        result["parameters"] = dictionary_parameters
        return result

    def split_message_by_tag(self, message: str) -> list[str]:
        """Splits a message by the control tags.

        :param `message`: The message to split."""
        return re.split(r"(<[^>]+>)", message)

    def generate_preset_msbp(self, name: str, project: MSBP) -> dict:
        """Generates an editable preset .py file for use with tag decoding mode of `preset`.

        :param `name`: Name for the file.
        :param `project`: A MSBP object."""

        with open(f"{name}.py", "w+") as preset:
            preset.write("from LMS.Stream.Reader import Reader\n")
            preset.write("from LMS.Stream.Writer import Writer\n\n")
            function_dictionary = 'preset = { "data": {'

            tag_groups = project.TGG2.groups[1:]

            for group_index, group in enumerate(tag_groups, start=1):
                group_name = group["name"]
                function_dictionary += (
                    f"\n\t{group_index}: {{\n\t'name': '{group_name}', \n\t'tags': [\n"
                )
                for tag_index in group["tag_indexes"]:
                    tag = project.TAG2.tags[tag_index]
                    tag_name = tag["name"]
                    read_function_name = f"read_{group_name}_{tag_name}"
                    write_function_name = f"write_{group_name}_{tag_name}"
                    function_dictionary += f"\t\t{{'name': '{tag_name}', 'read_function' : {read_function_name}, 'write_function': {write_function_name}}},\n"

                    preset.write(f"def {read_function_name}(reader: Reader):\n")
                    preset.write("\tdata = {}\n")

                    read_function = ""
                    write_function = ""
                    for parameter_index in tag["parameter_indexes"]:
                        parameter = project.TGP2.parameters[parameter_index]
                        parameter_name = parameter["name"]
                        parameter_type = parameter["type"]

                        if parameter_type in [
                            LMS_Types.uint8_0,
                            LMS_Types.uint8_1,
                            LMS_Types.float,
                        ]:
                            read_function += (
                                f"\tdata['{parameter_name}'] = reader.read_uint8()\n"
                            )
                            write_function += (
                                f"\twriter.write_uint8(int(data['{parameter_name}']))\n"
                            )

                        elif parameter_type in [
                            LMS_Types.uint16_0,
                            LMS_Types.uint16_1,
                            LMS_Types.uint16_2,
                        ]:
                            read_function += (
                                f"\tdata['{parameter_name}'] = reader.read_uint16()\n"
                            )
                            write_function += f"\twriter.write_uint16(int(data['{parameter_name}']))\n"

                        elif parameter_type in [LMS_Types.uint32_0, LMS_Types.uint32_1]:
                            read_function += (
                                f"\tdata['{parameter_name}'] = reader.read_uint32()\n"
                            )
                            write_function += f"\twriter.write_uint32(int(data['{parameter_name}']))\n"
                        elif parameter_type == LMS_Types.string:
                            read_function += f"\tdata['{parameter_name}'] = reader.read_len_prefixed_utf16_string()\n"
                            write_function += f"\twriter.write_len_prefixed_utf16_string(data['{parameter_name}'])\n"
                        else:
                            list_items = [
                                project.TGL2.items[item_index]
                                for item_index in parameter["item_indexes"]
                            ]
                            read_function += f"\tlist_items = {list_items}\n"
                            read_function += f"\tdata['{parameter_name}'] = list_items[reader.read_uint8()]\n"
                            write_function += f"\tvalue = data['{parameter_name}']\n"
                            write_function += f"\tlist_items = {list_items}\n"
                            write_function += (
                                f"\twriter.write_uint8(list_items.index(value))\n"
                            )

                        read_function += "\n"
                        write_function += "\n"

                    read_function += "\treturn data\n"
                    write_function += "\treturn\n"
                    preset.write(read_function + "\n\n")
                    preset.write(
                        f"def {write_function_name}(writer: Writer, data: dict):\n"
                    )
                    preset.write(write_function + "\n\n")

                function_dictionary += f"\t]}},"
            function_dictionary += "}}"
            preset.write(function_dictionary)

    def set_preset(self, preset: dict) -> None:
        self.preset = {}
        self.preset["data"] = {} 
        self.preset["data"] = base_preset["data"]

        for group in preset["data"]:
            self.preset["data"][group] = preset["data"][group]

        self.preset_set = True
            
    def read(self, reader: Reader, tag_decoding_mode: str = "default") -> None:
        """Reads the TXT2 block from a stream.

        :param `reader`: A Reader object.
        :param `tag_decoding_mode`: The mode at which to decode tags.
            * `default` uses near Kuriimu syntax `[n0.4:]`, `[n0.3:00-00-00-FF]`,
            * `preset` completely decodes it completely. `[System:PageBreak:]`, `[System:Color r="0" g="0" b="0" a="255]`.
                * A preset must be set using `TXT2.generate_preset_msbp` and set via `TXT2.set_preset`.
        """

        self.tag_decoding_mode = tag_decoding_mode

        if self.tag_decoding_mode != "default" and self.tag_decoding_mode != "preset":
            raise Exception(
                "An invalid tag decoding mode was provided, expected default or preset."
            )

        self.block.read_header(reader)
        message_count = reader.read_uint32()

        encoding = "UTF-16-LE" if reader.byte_order == "little" else "UTF-16-BE"
        tag_indicator = b"\x0E\x00" if reader.byte_order == "little" else b"\x00\x0E"

        # Read the messages
        offsets = self.block.get_item_offsets(reader, message_count)

        for i, offset in enumerate(offsets):
            if i < len(offsets) - 1:
                next_offset = offsets[i + 1]
            else:
                next_offset = self.block.data_start + self.block.size

            reader.seek(offset)
            message = b""

            while reader.tell() < next_offset:
                bytes = reader.read_bytes(2)
                if bytes == tag_indicator:
                    if tag_decoding_mode == "default":
                        message += self.read_encoded_control_tag(reader, encoding)
                        continue
                    message += self.read_decoded_control_tag(reader, encoding)
                else:
                    message += bytes

            self.messages.append(message.decode(encoding))

        self.block.seek_to_end(reader)

    def read_encoded_control_tag(self, reader: Reader, encoding: str) -> bytes:
        """Reads an encoded control tag from a stream.

        :param `reader`: A Reader object
        :param `encoding`: The encoding of the file."""
        group_index = reader.read_uint16()
        tag_index = reader.read_uint16()
        parameter_size = reader.read_uint16()
        end = reader.tell() + parameter_size

        if group_index in base_preset["data"]:
            group_name = base_preset["data"][group_index]["name"]
            tag_name = base_preset["data"][group_index]["tags"][tag_index]["name"]
            read_function = base_preset["data"][group_index]["tags"][tag_index][
                "read_function"
            ]
            data = read_function(reader)

            parsed_parameters = ""
            for parameter in data:
                parsed_parameters += f'{parameter}="{data[parameter]}" '

            tag = f"<{group_name}:{tag_name}: {parsed_parameters}".strip() + f">"
            reader.seek(end)
            return tag.encode(encoding)

        hex_parameters = reader.read_bytes(parameter_size).hex()
        encoded_parameters = "-".join(
            [
                hex_parameters[i] + hex_parameters[i + 1]
                for i in range(0, len(hex_parameters), 2)
            ]
        )

        return f"<n{group_index}.{tag_index}:{encoded_parameters}>".encode(encoding)

    def read_decoded_control_tag(self, reader: Reader, encoding: str) -> bytes:
        """Reads a decoded control tag from a stream.

        :param `reader`: A Reader object
        :param `encoding`: The encoding of the file."""

        start = reader.tell()
        group_index = reader.read_uint16()
        tag_index = reader.read_uint16()
        parameter_size = reader.read_uint16()
        parameter_start = reader.tell()

        end = parameter_start + parameter_size

        group_name = self.preset["data"][group_index]["name"]
        tag_name = self.preset["data"][group_index]["tags"][tag_index]["name"]
        read_function = self.preset["data"][group_index]["tags"][tag_index][
            "read_function"
        ]
        # Revert to unencoded tag if a tag isnt in the preset
        try:
            data = read_function(reader)
        except Exception:
            print(
                f"An occured while reading tag in function '{read_function.__name__}' parameter start at {parameter_start}, reverting to default."
            )
            reader.seek(start)
            return self.read_encoded_control_tag(reader, encoding)

        parsed_parameters = ""
        for parameter in data:
            parsed_parameters += f'{parameter}="{data[parameter]}" '

        tag = f"<{group_name}:{tag_name}: {parsed_parameters}".strip() + f">"
        reader.seek(end)
        return tag.encode(encoding)

    def write_encoded_control_tag(self, writer: Writer, tag: str) -> None:
        """Writes an encoded control tag to a stream.

        :param `reader`: A Reader object.
        :param `tag_data`: The tag."""
        group_index = tag[2 : tag.index(".")]
        tag_index = int(tag[tag.index(".") + 1 : tag.index(":")])

        parameters = tag[tag.rindex(":") + 1 : len(tag) - 1].split("-")
        parameter_size = len(parameters)

        # 1 indicates empty parameters due to split function returning an empty list
        if parameter_size == 1:
            parameter_size = 0

        writer.write_uint16(int(group_index))
        writer.write_uint16(int(tag_index))
        writer.write_uint16(parameter_size)

        for parameter in parameters:
            if parameter != "":
                writer.write_bytes(bytes.fromhex(parameter))

    def write_decoded_control_tag(self, writer: Writer, tag: str) -> None:
        """Writes a decoded control tag to a stream.

        :param `reader`: A Reader object.
        :param `tag_data`: The tag."""
        # TODO: Make this not manual and use a library
        has_parameters = tag[tag.rfind(":") + 1] != ">"
        decoded_parameters = {}
        if has_parameters:
            # Use this RE expression to avoid spaces in quotes cause im too lazy
            string_parameters = tag[tag.rfind(":") + 2 : len(tag) - 1]
            split_parameters = re.findall('(?:".*?"|\S)+', string_parameters)

            # Parse each individual parameter
            for parameter in split_parameters:
                name = parameter[0 : parameter.index("=")]
                value = parameter[parameter.index('"') + 1 : parameter.rindex('"')]
                decoded_parameters[name] = value

        group_name = tag[1 : tag.index(":")]
        tag_name = tag[tag.index(":") + 1 : tag.rindex(":")]

        # Start writing the tag data
        for group in self.preset["data"]:
            for tag_index, tag in enumerate(self.preset["data"][group]["tags"]):
                if (
                    group_name == self.preset["data"][group]["name"]
                    and tag["name"] == tag_name
                ):
                    writer.write_uint16(group)
                    writer.write_uint16(tag_index)

                    size_offset = writer.tell()
                    # Write 0 as placeholder for size
                    writer.write_uint16(0)
                    start = writer.tell()
                    # Run the write function preset
                    tag["write_function"](writer, decoded_parameters)
                    end = writer.tell()
                    size = end - start

        # Tags must be even size, padded with the padding char defined in preset
        # Only write the character if the character is provided, and write a null character
        # in case the user does not provide one and still wants to re-read the file.
        if size % 2 == 1:
            size += 1
            writer.seek(end)
            if "padding_char" in self.preset:
                writer.write_bytes(bytes.fromhex(self.preset["padding_char"]))
            else:
                writer.write_bytes(b"\x00")
            end += 1

        writer.seek(size_offset)
        writer.write_uint16(size)
        writer.seek(end)

    def write(self, writer: Writer) -> None:
        self.block.magic = "TXT2"
        self.block.size = 0
        self.block.write_header(writer)
        self.block.data_start = writer.tell()
        message_count = len(self.messages)
        writer.write_uint32(message_count)
        message_offset = message_count * 4 + 4

        in_tag = False
        encoded_messages = []
        for message in self.messages:
            encoded_message = b""
            message_writer = Writer(encoded_message, writer.byte_order)
            for i in range(len(message)):
                if message[i] == "<":
                    # Start parsing the tag (in a hacky way xD)
                    in_tag = True
                    start_index = i
                    end_index = i

                    # Get the end index of the tag
                    while True:
                        end_character = message[end_index]
                        if end_character == ">":
                            end_index += 1
                            break
                        end_index += 1

                    # Get the tag data
                    tag = message[start_index:end_index]
                    is_encoded = "." in tag[: tag.find(":")]

                    tag_indicator = (
                        b"\x0E\x00"
                        if message_writer.byte_order == "little"
                        else b"\x00\x0E"
                    )
                    message_writer.write_bytes(tag_indicator)

                    # Parse parameters
                    if is_encoded:
                        self.write_encoded_control_tag(message_writer, tag)
                        continue

                    self.write_decoded_control_tag(message_writer, tag)

                    i += end_index
                elif message[i] == ">":
                    in_tag = False
                else:
                    if in_tag:
                        continue
                    message_writer.write_utf16_string(message[i])

            encoded_messages.append(message_writer.get_data())

        size = message_count * 4 + 4
        for message in encoded_messages:
            writer.write_uint32(message_offset)
            message_offset += len(message)
            size += len(message)

        for message in encoded_messages:
            writer.write_bytes(message)

        self.block.size = size
        self.block.write_end_data(writer)



from LMS.Common.LMS_Block import LMS_Block
from LMS.Stream.Reader import Reader


class ALI2:
    """A class that represents a ALI2 block in a MSBP file.

    https://github.com/kinnay/Nintendo-File-Formats/wiki/MSBP-File-Format#ali2-block"""

    def __init__(self):
        self.block: LMS_Block = LMS_Block()
        self.attribute_lists: list[list] = []

    def read(self, reader: Reader) -> None:
        """Reads the ALI2 block from a stream.

        :param `reader`: A Reader object."""
        self.block.read_header(reader)

        list_count = reader.read_uint32()

        # Read the lists
        for list_offset in self.block.get_item_offsets(reader, list_count):
            list_items = []
            reader.seek(list_offset)
            item_count = reader.read_uint32()

            item_offsets = [
                list_offset + reader.read_uint32() for _ in range(item_count)
            ]

            for item_offset in item_offsets:
                reader.seek(item_offset)
                list_items.append(reader.read_string_nt())

            self.attribute_lists.append(list_items)

        self.block.seek_to_end(reader)



from LMS.Common.LMS_Block import LMS_Block
from LMS.Stream.Reader import Reader


class TGL2:
    """A class that represents a TGL2 block in a MSBT file.

    https://github.com/kinnay/Nintendo-File-Formats/wiki/MSBP-File-Format#tgl2-block"""

    def __init__(self):
        self.block: LMS_Block = LMS_Block()
        self.items: list[str] = []

    def read(self, reader: Reader) -> None:
        """Reads the TGL2 block from a stream.

        :param `reader`: A Reader object."""
        self.block.read_header(reader)

        item_count = reader.read_uint32()
        # Read the list items
        for offset in self.block.get_item_offsets(reader, item_count):
            reader.seek(offset)
            self.items.append(reader.read_string_nt())

        self.block.seek_to_end(reader)



from LMS.Common.LMS_Block import LMS_Block
from LMS.Common.LMS_Enum import LMS_Types
from LMS.Project.MSBP import MSBP
from LMS.Stream.Reader import Reader
from LMS.Stream.Writer import Writer


class ATR1:
    """A class that represents a ATR1 block in a MSBT file.

    https://github.com/kinnay/Nintendo-File-Formats/wiki/MSBT-File-Format#atr1-block"""

    def __init__(self):
        self.block: LMS_Block = LMS_Block()
        self.attributes: list[dict | bytes] = []
        self.structure: dict = {}

    def create_decoded_attribute(self) -> dict:
        """Adds an attribute, setting each value to the default for each type."""
        attribute = {}
        for label in self.structure:
            type = self.structure[label]["type"]
            match type:
                case LMS_Types.uint8_0:
                    attribute[label] = 0
                case LMS_Types.uint8_1:
                    attribute[label] = 0
                case LMS_Types.float:
                    attribute[label] = 0
                case LMS_Types.uint16_0:
                    attribute[label] = 0
                case LMS_Types.uint16_1:
                    attribute[label] = 0
                case LMS_Types.uint16_2:
                    attribute[label] = 0
                case LMS_Types.uint32_0:
                    attribute[label] = 0
                case LMS_Types.uint32_1:
                    attribute[label] = 0
                case LMS_Types.string:
                    attribute[label] = ""
                case LMS_Types.list_index:
                    attribute[label] = self.structure[label]["list_items"][0]
        return attribute

    def read(self, reader: Reader, project: MSBP = None) -> None:
        """Reads the ATR1 block from a stream.

        :param `reader`: A Reader object.
        :param `project`: A MSBP object used for decoding attributes."""

        self.block.read_header(reader)
        attribute_count = reader.read_uint32()
        bytes_per_attribute = reader.read_uint32()

        # Read the attributes
        if not isinstance(project, MSBP) or len(project.ALB1.labels) == 0:
            self.attributes = [reader.read_bytes(bytes_per_attribute) for _ in range(attribute_count)]
            return
        
        # Create the structure
        for index in project.ALB1.labels:
            label = project.ALB1.labels[index]
            type = project.ATI2.attributes[index]["type"]
            offset = project.ATI2.attributes[index]["offset"]
            list_index = project.ATI2.attributes[index]["list_index"]

            self.structure[label] = {
                "type": type,
                "offset": offset,
                "list_index": list_index,
            }

            if type == LMS_Types.list_index:
                self.structure[label]["list_items"] = project.ALI2.attribute_lists[
                    list_index
                ]

        # Write the attributes
        for _ in range(attribute_count):
            attribute = {}
            for label in self.structure:
                type = self.structure[label]["type"]
                match type:
                    case LMS_Types.uint8_0:
                        attribute[label] = reader.read_uint8()
                    case LMS_Types.uint8_1:
                        attribute[label] = reader.read_uint8()
                    case LMS_Types.float:
                        attribute[label] = reader.read_uint8()
                    case LMS_Types.uint16_0:
                        attribute[label] = reader.read_uint16()
                    case LMS_Types.uint16_1:
                        attribute[label] = reader.read_uint16()
                    case LMS_Types.uint16_2:
                        attribute[label] = reader.read_uint16()
                    case LMS_Types.uint32_0:
                        attribute[label] = reader.read_uint32()
                    case LMS_Types.uint32_1:
                        attribute[label] = reader.read_uint32()
                    case LMS_Types.string:
                        offset = self.block.data_start + reader.read_uint32()
                        last = reader.tell()
                        reader.seek(offset)
                        string = reader.read_utf16_string()
                        attribute[label] = string
                        reader.seek(last)
                    case LMS_Types.list_index:
                        index = reader.read_uint8()
                        attribute[label] = self.structure[label]["list_items"][index]

            self.attributes.append(attribute)

        self.block.seek_to_end(reader)

    def write(self, writer: Writer):
        """Writes the ATR1 block to a stream.

        :param `reader`: A Reader object.
        """
        self.block.magic = "ATR1"
        self.block.write_header(writer)
        self.block.data_start = writer.tell()

        bytes_per_attribute = 0
        attribute_count = len(self.attributes)
        writer.write_uint32(attribute_count)

        # If an error occurs trying to get length, there are no attributes
        try:
            attribute = self.attributes[0]
        except IndexError:
            attribute = None 
            bytes_per_attribute = 0
     
        if isinstance(attribute, bytes):
            bytes_per_attribute = len(attribute)
        else:
            for label in self.structure:
                type = self.structure[label]["type"]
                match type:
                    case LMS_Types.uint8_0:
                        bytes_per_attribute += 1
                    case LMS_Types.uint8_1:
                        bytes_per_attribute += 1
                    case LMS_Types.float:
                        bytes_per_attribute += 1
                    case LMS_Types.uint16_0:
                        bytes_per_attribute += 2
                    case LMS_Types.uint16_1:
                        bytes_per_attribute += 2
                    case LMS_Types.uint16_2:
                        bytes_per_attribute += 2
                    case LMS_Types.uint32_0:
                        bytes_per_attribute += 4
                    case LMS_Types.uint32_1:
                        bytes_per_attribute += 4
                    case LMS_Types.string:
                        bytes_per_attribute += 4
                    case LMS_Types.list_index:
                        bytes_per_attribute += 1

        writer.write_uint32(bytes_per_attribute)

        # Verify each attribute is type bytes meaning to write the raw parameters
        if all(isinstance(attribute, bytes) for attribute in self.attributes):
            for attribute in self.attributes:
                writer.write_bytes(attribute)

            self.block.size = 8 + attribute_count * bytes_per_attribute
            self.block.write_end_data(writer)
            return

        # Get the bytes_per_attribute
        strings = []
        string_offset = 8 + attribute_count * bytes_per_attribute
        for attribute in self.attributes:
            for label in self.structure:
                type = self.structure[label]["type"]
                match type:
                    case LMS_Types.uint8_0:
                        writer.write_uint8(attribute[label])
                    case LMS_Types.uint8_1:
                        writer.write_uint8(attribute[label])
                    case LMS_Types.float:
                        writer.write_uint8(attribute[label])
                    case LMS_Types.uint16_0:
                        writer.write_uint16(attribute[label])
                    case LMS_Types.uint16_1:
                        writer.write_uint16(attribute[label])
                    case LMS_Types.uint16_2:
                        writer.write_uint16(attribute[label])
                    case LMS_Types.uint32_0:
                        writer.write_uint32(attribute[label])
                    case LMS_Types.uint32_1:
                        writer.write_uint32(attribute[label])
                    case LMS_Types.string:
                        string = attribute[label]
                        strings.append(string)
                        writer.write_uint32(string_offset)
                        string_offset += (
                            len(string.encode(writer.get_utf16_encoding())) + 2
                        )
                    case LMS_Types.list_index:
                        writer.write_uint8(
                            self.structure[label]["list_items"].index(attribute[label])
                        )

        string_size = 8
        for string in strings:
            writer.write_utf16_string(string, use_double=True)
            string_size += len(string.encode(writer.get_utf16_encoding())) + 2

        self.block.size = attribute_count * bytes_per_attribute + string_size
        self.block.write_end_data(writer)



import struct
from io import BytesIO

types = {
    "little": {"uint8": "<B", "uint16": "<H", "uint32": "<I"},
    "big": {"uint8": ">B", "uint16": ">H", "uint32": ">I"},
}


class Reader:
    """A class utilized for reading binary files into a stream object."""

    def __init__(self, data: bytes, byte_order: str = "little"):
        self.data = BytesIO(data)
        self.byte_order = byte_order

    def change_byte_order(self, byte_order: str) -> None:
        """Changes the byte order of the Reader."""
        self.byte_order = byte_order

    def get_utf16_encoding(self) -> str:
        """Returns the UTF-16 encoding of the stream."""
        return "UTF-16-LE" if self.byte_order == "little" else "UTF-16-BE"

    def skip(self, length: int) -> None:
        """Skips `length` amount of bytes."""
        self.data.read(length)

    def read_bytes(self, length: int) -> bytes:
        """Reads `length` amount of bytes."""
        return self.data.read(length)

    def seek(self, offset: int, whence: int = 0) -> None:
        """Seeks to an offset with whence."""
        self.data.seek(offset, whence)

    def tell(self) -> int:
        """Returns the current position in the stream."""
        return self.data.tell()

    def read_uint8(self) -> int:
        """Reads a UInt8 from the stream."""
        return struct.unpack(types[self.byte_order]["uint8"], self.data.read(1))[0]

    def read_uint16(self) -> int:
        """Reads a UInt16 from the stream."""
        return struct.unpack(types[self.byte_order]["uint16"], self.data.read(2))[0]

    def read_uint32(self) -> int:
        """Reads a UInt32 from the stream."""
        return struct.unpack(types[self.byte_order]["uint32"], self.data.read(4))[0]

    def read_string_len(self, length: int) -> str:
        """Reads a string that is length bytes long from the stream."""
        return self.data.read(length).decode("UTF-8")

    def read_string_nt(self) -> str:
        """Reads a null terminated string from the stream."""
        result = b""
        char = self.data.read(1)
        while char != b"\x00":
            result += char
            char = self.data.read(1)
        return result.decode("UTF-8")

    def read_utf16_string(self):
        """Reads a UTF-16 string from the stream."""
        mode = "UTF-16-LE" if self.byte_order == "little" else "UTF-16-BE"

        message = b""
        byte = self.data.read(2)
        while True:
            if byte == b"\x00\x00":
                break
            message += byte
            byte = self.data.read(2)

        return message.decode(mode)

    def read_len_prefixed_utf16_string(self):
        """Reads a UTF-16 string from the stream."""
        length = self.read_uint16()
        string = self.read_bytes(length).decode("UTF-16-LE")
        return string



from enum import Enum


class LMS_MessageEncoding(Enum):
    """An enum for representing message encoding. As in offical LMS files."""
    UTF8 = 0
    UTF16 = 1
    UTF32 = 2

class LMS_Types(Enum):
    """A class that represents binary types utilized in attributes and tags."""
    uint8_0 = 0
    uint8_1 = 3
    float = 6 # According to Trippixyz on discord, float will be parsed like a regular UInt8 though.
    uint16_0 = 1
    uint16_1 = 4
    uint16_2 = 7
    uint32_0 = 2
    uint32_1 = 5
    string = 8
    list_index = 9



from LMS.Common.LMS_Block import LMS_Block
from LMS.Stream.Reader import Reader
from LMS.Common.LMS_Enum import LMS_Types


class TGP2:
    """A class that represents a TGP2 block in a MSBT file.

    https://github.com/kinnay/Nintendo-File-Formats/wiki/MSBP-File-Format#tgp2-block"""

    def __init__(self):
        self.block: LMS_Block = LMS_Block()
        self.parameters: list[dict] = []

    def read(self, reader: Reader) -> None:
        """Reads the TGP2 block from a stream.

        :param `reader`: A Reader object."""
        self.block.read_header(reader)

        parameter_count = reader.read_uint16()
        reader.skip(2)
        # Read the parameters
        for offset in self.block.get_item_offsets(reader, parameter_count):
            parameter = {}
            reader.seek(offset)
            type = LMS_Types(reader.read_uint8())
            parameter["type"] = type

            if type != LMS_Types.list_index:
                parameter["name"] = reader.read_string_nt()
                self.parameters.append(parameter)
                continue

            reader.skip(1)
            list_count = reader.read_uint16()
            parameter["item_indexes"] = [
                reader.read_uint16() for _ in range(list_count)
            ]
            parameter["name"] = reader.read_string_nt()

            self.parameters.append(parameter)

        self.block.seek_to_end(reader)



from LMS.Common.LMS_Block import LMS_Block
from LMS.Common.LMS_Enum import LMS_Types
from LMS.Stream.Reader import Reader


class ATI2:
    """A class that represents a ATI2 block in a MSBP file.

    https://github.com/kinnay/Nintendo-File-Formats/wiki/MSBP-File-Format#ati2-block"""

    def __init__(self):
        self.block: LMS_Block = LMS_Block()
        self.attributes: list[dict] = []

    def read(self, reader: Reader) -> None:
        """Reads the ATI2 block from a stream.

        :param `reader`: A Reader object."""
        self.block.read_header(reader)

        attribute_count = reader.read_uint32()
        # Read the attribute information
        for _ in range(attribute_count):
            attribute = {}
            attribute["type"] = LMS_Types(reader.read_uint8())
            reader.skip(1)
            attribute["list_index"] = reader.read_uint16()
            attribute["offset"] = reader.read_uint32()
            self.attributes.append(attribute)

        self.block.seek_to_end(reader)



from LMS.Common.LMS_Binary import LMS_Binary
from LMS.Common.LMS_Enum import LMS_MessageEncoding
from LMS.Project.MSBP import MSBP

from LMS.Stream.Reader import Reader
from LMS.Stream.Writer import Writer

from LMS.Common.LMS_HashTable import LMS_HashTable
from LMS.Message.TXT2 import TXT2
from LMS.Message.ATR1 import ATR1



class MSBT:
    """A class that represents a Message Studio Binary Text file.

    https://github.com/kinnay/Nintendo-File-Formats/wiki/MSBT-File-Format"""

    def __init__(self):
        self.binary: LMS_Binary = LMS_Binary()
        self.LBL1: LMS_HashTable = LMS_HashTable()
        self.ATR1: ATR1 = ATR1()
        self.TXT2: TXT2 = TXT2()

    def read(
        self, reader: Reader, msbp: MSBP = None, tag_decoding_mode: str = "default"
    ) -> None:
        """Reads a MSBT file from a stream.

        :param `reader`: A Reader object.
        :param `msbp`: A MSBP object. Used for decoding of attributes.
        :param `tag_decoding_mode`: The mode at which to decode tags.
            * `default` uses near Kuriimu syntax `<n0.4:>`, `<n0.3:00-00-00-FF>`,
            * `preset` completely decodes it `<System:PageBreak:>`, `<System:Color r="0" g="0" b="0" a="255">`.
                * A preset must be set using `TXT2.generate_preset_msbp` or a manual one that has been created.
        """
        self.binary.read_header(reader)

        lbl1_valid, lbl1_offset = self.binary.search_block_by_name(reader, "LBL1")
        atr1_valid, atr1_offset = self.binary.search_block_by_name(reader, "ATR1")
        txt2_valid, txt2_offset = self.binary.search_block_by_name(reader, "TXT2")
       
        # Read LBL1
        if lbl1_valid:
            reader.seek(lbl1_offset)
            self.LBL1.read(reader)
        else:
            self.LBL1 = None

        # Read ATR1
        if atr1_valid:
            reader.seek(atr1_offset)
            self.ATR1.read(reader, msbp)
        else:
            self.ATR1 = None

        # Read TXT2
        if txt2_valid:
            reader.seek(txt2_offset)
            self.TXT2.read(reader, tag_decoding_mode)
        else:
            self.TXT2 = None

    def write(self, writer: Writer) -> None:
        """Writes a MSBT file to a stream.

        :param `reader`: A Reader object."""
        block_count = 0

        if self.LBL1 is not None:
            block_count += 1

        if self.ATR1 is not None:
            block_count += 1
        
        if self.TXT2 is not None:
            block_count += 1

        self.binary.magic = "MsgStdBn"
        self.binary.encoding = LMS_MessageEncoding(2)
        self.binary.revision = 3
        self.binary.block_count = block_count

        self.binary.write_header(writer)
        self.LBL1.block.magic = "LBL1"
        self.ATR1.block.magic = "LBL1"
        self.TXT2.block.magic = "TXT2"

        if self.LBL1 is not None:
            self.LBL1.write(writer, 101)

        if self.ATR1 is not None:
            self.ATR1.write(writer)

        if self.TXT2 is not None:
            self.TXT2.write(writer)

        writer.seek(0, 2)
        size = writer.tell()
        writer.seek(18)
        writer.write_uint32(size)
